package com.jm3.scene;

import com.jme3.math.Vector3f;
import com.jme3.scene.Mesh;
import com.jme3.scene.VertexBuffer;
import com.jme3.util.BufferUtils;
import java.util.ArrayList;

/**
 * Class that implements terrain generating algorithms.
 *
 * @author G. Amador & A. Gomes
 */
public class Terrain {

    private ArrayList<ArrayList<Vector3f>> Heightfield;
    private float cellSideSize;
    private final int dimension;
    private final float roughness;
    private float min, max;

    /**
     * Constructor, sets the application Window, and initializes the terrain
     * height map. Used for diamond square.
     *
     * @param lod the level of detail of the heightmap to generate. lod
     * corresponds to the integer variable k in the formula (2 ^ k) + 1, where k
     * = 1, 2, 3 ... is the value of both the length and width of the heightmap,
     * stored in the TerrainGen class "terrain" variable. The value of (2 ^ k) +
     * 1 gives the number of points per line of the grid. The higher k is the
     * more detailed is the generated heightmap, i.e., more points are
     * generated. Consequently, more details means more time to calculate the
     * heightmap values, and to render the generated terrain.
     * @param roughness the value that will increase or decrease the random
     * interval added to each new generated height.
     * @param size (considering a squared game world) of one of the game world
     * sides.
     */
    public Terrain(int lod, float roughness, float size) {
        this.dimension = (int) Math.pow(2, lod) + 1;
        this.roughness = roughness;
        cellSideSize = size / dimension;
        resetTerrain();
    }

    /**
     * This method returns the terrain generated with the diamond square
     * algorithm.
     *
     * @return the last terrain generated with the diamond square algorithm.
     */
    public ArrayList<Mesh> getTerrain() {
        ArrayList<Mesh> meshesTerrain = new ArrayList<>();
        for (int i = 0; i < (Heightfield.size() - 1); i++) {
            for (int j = 0; j < (Heightfield.get(i).size() - 1); j++) {
                Mesh meshTerrain = new Mesh();
                meshTerrain.setMode(Mesh.Mode.Lines);

                Vector3f[] verticesTerrain = new Vector3f[5];
                int[] indexes = new int[2 * verticesTerrain.length]; //Indexes are in pairs, from a vertex and to a vertex
                for (int k = 0; k < verticesTerrain.length - 1; k++) {
                    indexes[2 * k] = k;
                    indexes[2 * k + 1] = (k + 1);
                }

                verticesTerrain[0] = new Vector3f(
                        Heightfield.get(i).get(j).x,
                        Heightfield.get(i).get(j).y,
                        Heightfield.get(i).get(j).z);

                verticesTerrain[1] = new Vector3f(
                        Heightfield.get(i).get(j + 1).x,
                        Heightfield.get(i).get(j + 1).y,
                        Heightfield.get(i).get(j + 1).z);

                verticesTerrain[2] = new Vector3f(
                        Heightfield.get(i + 1).get(j + 1).x,
                        Heightfield.get(i + 1).get(j + 1).y,
                        Heightfield.get(i + 1).get(j + 1).z);

                verticesTerrain[3] = new Vector3f(
                        Heightfield.get(i + 1).get(j).x,
                        Heightfield.get(i + 1).get(j).y,
                        Heightfield.get(i + 1).get(j).z);

                verticesTerrain[4] = new Vector3f(
                        Heightfield.get(i).get(j).x,
                        Heightfield.get(i).get(j).y,
                        Heightfield.get(i).get(j).z);

                /* Setting the Meshes Buffer */
                meshTerrain.setBuffer(VertexBuffer.Type.Position, 3, BufferUtils.createFloatBuffer(verticesTerrain));
                meshTerrain.setBuffer(VertexBuffer.Type.Index, 2, BufferUtils.createIntBuffer(indexes));
                meshTerrain.updateBound();
                meshTerrain.updateCounts();

                meshesTerrain.add(meshTerrain);
            }
        }

        return meshesTerrain;
    }

    /**
     * This method returns the terrain generated with the diamond square
     * algorithm.
     */
    public void generateTerrain() {
        resetTerrain();
        Heightfield = diamondSquare();
    }

    /**
     * Set all values of the height map to zero.
     */
    private void resetTerrain() {
        Heightfield = new ArrayList<>();
        float x_pos = 0;
        for (int x = 0; x < dimension; x++, x_pos += cellSideSize) {
            ArrayList<Vector3f> line = new ArrayList<>(dimension);
            float z_pos = 0;
            for (int z = 0; z < dimension; z++, z_pos += cellSideSize) {
                Vector3f v = new Vector3f(x_pos, 0, z_pos);
                line.add(v);
            }
            Heightfield.add(line);
        }
    }

    /**
     * Terrain generation using the diamond square algorithm.
     *
     * @return A Heightfield stored in an ArrayList<ArrayList<Vector3f>>.
     */
    private ArrayList<ArrayList<Vector3f>> diamondSquare() {
        float rn = roughness;
        int size = dimension;

        /**
         * BEGIN OF EXERCISE 2 (Step 2): complete the diamond square algorithm
         * in the following TODO's for Step 2.
         */
        // TODO: (Sub-Step 2.1) ADD YOUR CODE HERE!
        //for ... [dimension - 1; 2] taking a step of last step lenght / 2 
        //  generate the new square values
        //  generate the new diamond values        
        
        
        // END OF EXERCISE 2 (Step 2). -----------------------------------------  
        
        return Heightfield;
    }
}
